package PVE::QemuServer;

use strict;
use POSIX;
use IO::Handle;
use IO::Select;
use IO::File;
use IO::Dir;
use IO::Socket::UNIX;
use File::Basename;
use File::Path;
use File::stat;
use Getopt::Long;
use Digest::SHA1;
use Fcntl ':flock';
use Cwd 'abs_path';
use IPC::Open3;
use Fcntl;
use Sys::Syslog;
use PVE::Storage;

# Note about locking: we use flock on the config file protect 
# against concurent actions.
# Adiitionaly, we have a 'lock' setting in the config file. This
# can be set to 'migrate' or 'backup'. Most actions are not
# allowed when such lock is set. But you can ignore this kind of
# lock wth the --skiplock flag.

#no warnings 'redefine';

unless(defined(&_VZSYSCALLS_H_)) {
    eval 'sub _VZSYSCALLS_H_ () {1;}' unless defined(&_VZSYSCALLS_H_);
    require 'sys/syscall.ph';
    if(defined(&__x86_64__)) {
	eval 'sub __NR_fairsched_vcpus () {499;}' unless defined(&__NR_fairsched_vcpus);
	eval 'sub __NR_fairsched_mknod () {504;}' unless defined(&__NR_fairsched_mknod);
	eval 'sub __NR_fairsched_rmnod () {505;}' unless defined(&__NR_fairsched_rmnod);
	eval 'sub __NR_fairsched_chwt () {506;}' unless defined(&__NR_fairsched_chwt);
	eval 'sub __NR_fairsched_mvpr () {507;}' unless defined(&__NR_fairsched_mvpr);
	eval 'sub __NR_fairsched_rate () {508;}' unless defined(&__NR_fairsched_rate);
	eval 'sub __NR_setluid () {501;}' unless defined(&__NR_setluid);
	eval 'sub __NR_setublimit () {502;}' unless defined(&__NR_setublimit);
    }
    elsif(defined( &__i386__) ) {
	eval 'sub __NR_fairsched_mknod () {500;}' unless defined(&__NR_fairsched_mknod);
	eval 'sub __NR_fairsched_rmnod () {501;}' unless defined(&__NR_fairsched_rmnod);
	eval 'sub __NR_fairsched_chwt () {502;}' unless defined(&__NR_fairsched_chwt);
	eval 'sub __NR_fairsched_mvpr () {503;}' unless defined(&__NR_fairsched_mvpr);
	eval 'sub __NR_fairsched_rate () {504;}' unless defined(&__NR_fairsched_rate);
	eval 'sub __NR_fairsched_vcpus () {505;}' unless defined(&__NR_fairsched_vcpus);
	eval 'sub __NR_setluid () {511;}' unless defined(&__NR_setluid);
	eval 'sub __NR_setublimit () {512;}' unless defined(&__NR_setublimit);
    } else {
	die("no fairsched syscall for this arch");
    }
    require 'asm/ioctl.ph';
    eval 'sub KVM_GET_API_VERSION () { &_IO(0xAE, 0x);}' unless defined(&KVM_GET_API_VERSION);
}

sub fairsched_mknod {
    my ($parent, $weight, $desired) = @_;

    return syscall(&__NR_fairsched_mknod, int ($parent), int ($weight), int ($desired));
}

sub fairsched_rmnod {
    my ($id) = @_;

    return syscall(&__NR_fairsched_rmnod, int ($id));
}

sub fairsched_mvpr {
    my ($pid, $newid) = @_;

    return syscall(&__NR_fairsched_mvpr, int ($pid), int ($newid));
}

sub fairsched_vcpus {
    my ($id, $vcpus) = @_;

    return syscall(&__NR_fairsched_vcpus, int ($id), int ($vcpus));
}

sub fairsched_rate {
    my ($id, $op, $rate) = @_;

    return syscall(&__NR_fairsched_rate, int ($id), int ($op), int ($rate));
}

use constant FAIRSCHED_SET_RATE  => 0;
use constant FAIRSCHED_DROP_RATE => 1;
use constant FAIRSCHED_GET_RATE  => 2;

sub fairsched_cpulimit {
    my ($id, $limit) = @_;

    my $cpulim1024 = int ($limit * 1024 / 100);
    my $op = $cpulim1024 ? FAIRSCHED_SET_RATE : FAIRSCHED_DROP_RATE;

    return fairsched_rate ($id, $op, $cpulim1024);
}

my $confdir = '/etc/qemu-server';

my $var_run_tmpdir = "/var/run/qemu-server";
my $lock_dir = "/var/lock/qemu-server";

my $pcisysfs = "/sys/bus/pci";

my $default_memory = 512;

my $confvars = {
    onboot => 'bool',
    autostart => 'bool',
    reboot => 'bool',
    lock => 'lockinfo',
    cpulimit => 'natural',
    cpuunits => 'natural',
    memory => 'natural',
    keyboard => 'lang',
    name => 'string',
    description => 'string',
    ostype => 'ostype',
    boot => 'boot',
    bootdisk => 'bootdisk',
    smp => 'natural',
    sockets => 'natural',
    cores => 'natural',
    acpi => 'bool',
    kvm => 'bool',
    tdf => 'bool',
    localtime => 'bool',
    freeze => 'bool',
    vlanu => 'vlan',
    vga => 'vgatype',
    hostusb => 'hostusb',
    hostpci => 'hostpci',
    serial => 'serial',
    parallel => 'parallel',
    startdate => 'startdate',
    args => 'string',
    tablet => 'bool',
    migrate_speed => 'natural',
    migrate_downtime => 'natural',

    #cpu => 'string',
    #machine => 'string',
    #fda => 'file',
    #fdb => 'file',
    #mtdblock => 'file',
    #sd => 'file',
    #pflash => 'file',
    #snapshot => 'bool',
    #bootp => 'file',
    ##tftp => 'dir',
    ##smb => 'dir',
    ##usb
    ##usbdevice
    #kernel => 'file',
    #append => 'string',
    #initrd => 'file',

    ##soundhw => 'string',
};

my $MAX_IDE_DISKS = 4;
my $MAX_SCSI_DISKS = 16;
my $MAX_VIRTIO_DISKS = 16;
my $MAX_VLANS = 4095;

for (my $i = 0; $i < $MAX_VLANS; $i++)  {
    $confvars->{"vlan$i"} = 'vlan';
}

my $drivename_hash;

for (my $i = 0; $i < $MAX_IDE_DISKS; $i++)  {
    $confvars->{"ide$i"} = 'drive';
    $drivename_hash->{"ide$i"} = 1;
}
for (my $i = 0; $i < $MAX_SCSI_DISKS; $i++)  {
    $confvars->{"scsi$i"} = 'drive';
    $drivename_hash->{"scsi$i"} = 1;
}
for (my $i = 0; $i < $MAX_VIRTIO_DISKS; $i++)  {
    $confvars->{"virtio$i"} = 'drive';
    $drivename_hash->{"virtio$i"} = 1;
}

my $defvars = {
    onboot => ['bool', 0],
    autostart => ['bool', 0],
    keyboard => ['lang', undef],
    cpuunits => ['natural', 1000],
    memory => ['natural', 512],
    vga => ['vgatype', 'cirrus'],
    tdf => ['bool', 1],
    tablet => ['bool', 1],
    migrate_speed => ['natural', 0],
    migrate_downtime => ['natural', 1],
};

my $kvm_api_version = 0;

sub kvm_version {

    return $kvm_api_version if $kvm_api_version;

    my $fh = IO::File->new ("</dev/kvm") ||
	return 0;

    if (my $v = $fh->ioctl (KVM_GET_API_VERSION(), 0)) {
	$kvm_api_version = $v;
    }

    $fh->close();

    return  $kvm_api_version;
}

my $kvm_user_version;

sub kvm_user_version {

    return $kvm_user_version if $kvm_user_version;

    $kvm_user_version = 'unknown';

    my $tmp = `kvm -help 2>/dev/null`;
    
    if ($tmp =~ m/^QEMU( PC)? emulator version (\d+\.\d+\.\d+) /) {
	$kvm_user_version = $2;
    }

    return $kvm_user_version;

}

my $kernel_has_vhost_net = -c '/dev/vhost-net';

sub disknames {
    # order is important - used to autoselect boot disk
    return ((map { "ide$_" } (0 .. ($MAX_IDE_DISKS - 1))),  
	    (map { "scsi$_" } (0 .. ($MAX_SCSI_DISKS - 1))),
	    (map { "virtio$_" } (0 .. ($MAX_VIRTIO_DISKS - 1))));
}

sub valid_drivename {
    my $dev = shift;

    return defined ($drivename_hash->{$dev});
}

sub nic_models {
    return ['rtl8139', 'ne2k_pci', 'e1000',  'pcnet',  'virtio',
	    'ne2k_isa', 'i82551', 'i82557b', 'i82559er'];
}

sub os_list_description {

    return {
	other => 'Other',
	wxp => 'Windows XP',
	w2k => 'Windows 2000',
	w2k3 =>, 'Windows 2003',
	w2k8 => 'Windows 2008',
	wvista => 'Windows Vista',
	l24 => 'Linux 2.4',
	l26 => 'Linux 2.6',
    }; 
}

# a clumsy way to split an argument string into an array,
# we simply pass it to the cli (exec call)
sub split_args {
    my ($str) = @_;

    my $args = [];

    return $args if !$str;

    my $cmd = 'perl -e \'foreach my $a (@ARGV) { print "$a\n"; } \' -- ' . $str;

    eval {
	PVE::Storage::run_command ($cmd, outfunc => sub {
	    my $data = shift;
	    push @$args, $data;
	});
    };
    
    my $err = $@;

    die "unable to parse args: $str\n" if $err;

    return $args;
}

sub disk_devive_info {
    my $dev = shift;

    die "unknown disk device format '$dev'" if $dev !~ m/^(ide|scsi|virtio)(\d+)$/;

    my $bus = $1;
    my $index = $2;
    my $maxdev = 1024;

    if ($bus eq 'ide') {
	$maxdev = 2;
    } elsif ($bus eq 'scsi') {
	$maxdev = 8;
    }

    my $controller = int ($index / $maxdev);
    my $unit = $index % $maxdev;


    return { bus => $bus, desc => uc($bus) . " $controller:$unit",
	     controller => $controller, unit => $unit, index => $index };

}

sub qemu_drive_name {
    my ($dev, $media) = @_; 

    my $info = disk_devive_info ($dev);
    my $mediastr = '';

    if (($info->{bus} eq 'ide') || ($info->{bus} eq 'scsi')) {
	$mediastr = ($media eq 'cdrom') ? "-cd" : "-hd";
	return sprintf("%s%i%s%i", $info->{bus}, $info->{controller}, 
		       $mediastr, $info->{unit});
    } else {
	return sprintf("%s%i", $info->{bus}, $info->{index}); 
    }
}

sub new {
    my ($type, $storecfg) = @_;

    my $self = {};

    my $class = ref ($type) || $type;

    bless ($self, $class);

    $storecfg = PVE::Storage::load_config() if !$storecfg;

    $self->{storecfg} = $storecfg;

    return $self;
}

my $cdrom_path;

sub get_cdrom_path {

    return  $cdrom_path if $cdrom_path;

    return $cdrom_path = "/dev/cdrom" if -l "/dev/cdrom";
    return $cdrom_path = "/dev/cdrom1" if -l "/dev/cdrom1";
    return $cdrom_path = "/dev/cdrom2" if -l "/dev/cdrom2";
}

sub get_iso_path {
    my ($storecfg, $vmid, $cdrom) = @_;

    if ($cdrom eq 'cdrom') {
	return get_cdrom_path();
    } elsif ($cdrom eq 'none') {
	return '';
    } elsif ($cdrom =~ m|^/|) {
	return $cdrom;
    } else {
	return PVE::Storage::path ($storecfg, $cdrom);
    }
}

# try to convert old style file names to volume IDs
sub filename_to_volume_id {
    my ($vmid, $file, $media) = @_;

    if (!($file eq 'none' || $file eq 'cdrom' ||
	  $file =~ m|^/dev/.+| || $file =~ m/^([^:]+):(.+)$/)) {
			    
	return undef if $file =~ m|/|;
	
	if ($media && $media eq 'cdrom') {
	    $file = "local:iso/$file";
	} else {
	    $file = "local:$vmid/$file";
	}
    }

    return $file;
}

sub parse_options {
    my ($self, $vmid, $create) = @_;

    my $settings = {};

    my $get_option_func = sub {
	my ($opt, $value) = @_;
	    
	if ($opt eq 'cdrom') {
	    $opt = 'ide2';
	    if ($value && $value ne 'undef' && $value ne 'delete' && 
		$value !~ m/media=/) {
		$value .= ",media=cdrom"; 
	    }
	}

	my $ct = $confvars->{$opt};

	my $v;
	if (defined ($v = check_type ($vmid, $ct, $opt, $value))) {
	    my $drive = parse_drive ($opt, $value);
	    if (($ct eq 'drive') &&
		($drive->{file} !~ m/^(undef|delete|cdrom|none)$/) &&
		($drive->{file} !~ m|^/dev/.+|) &&
		($drive->{file} !~ m/^([^:]+):(.+)$/) &&
		($drive->{file} !~ m/^\d+$/)) { 
		# try to convert path to volume id
		my $media = $drive->{media} || 'disk';
		my ($vtype, $volid) = PVE::Storage::path_to_volume_id 
		    ($self->{storecfg}, $drive->{file});
		if (!$vtype) {
		    die "unable to parse value for '$opt'\n";
		} elsif ($vtype eq $media) {
		    $drive->{file} = $volid;
		    $v = print_drive ($vmid, $drive);
		} else {
		    die "unexpected media type for '$opt' ($vtype != $media)\n";
		}
	    }
	    $settings->{$opt} = $v;
	} else {
	    if (!$create && ($value eq '' || $value eq 'undef')) {
		$settings->{$opt} = $value;
	    } else {
		die "unable to parse value for '$opt'\n";
	    }
	}
    };

    my $getopt;
    foreach my $opt (keys %$confvars) {
	$getopt->{"$opt=s"} = $get_option_func;
    }

    $getopt->{"cdrom=s"} = $get_option_func;
 
    if (!GetOptions (%$getopt)) {
	return undef;
    }

    return $settings;
}

#sub create_conf {
#    my ($vmid, $settings) = @_;
#
#    lock_config ($vmid, &create_conf_nolock, $settings);
#}

sub create_conf_nolock {
    my ($self, $vmid, $settings) = @_;

    my $filename = config_file ($vmid);

    die "configuration file '$filename' already exists\n" if -f $filename;
    
    my $defaults = load_defaults();

    $settings->{name} = "vm$vmid" if !$settings->{name};
    if (!$settings->{memory}) {
	$settings->{memory} = $defaults->{memory} || $default_memory;
    }

    my $fh = new IO::File ($filename, "w") ||
	die "unable to create config for VM $vmid - $!\n";
    
    foreach my $opt (keys %$settings) {
	next if !$confvars->{$opt};

	my $value = $settings->{$opt};
	next if !$value;

	print $fh "$opt: $value\n";
    }

    $fh->close();
}

# ideX = [volume=]volume-id[,media=d][,cyls=c,heads=h,secs=s[,trans=t]]
#        [,snapshot=on|off][,cache=on|off][,format=f][,backup=yes|no]

sub parse_drive {
    my ($key, $data) = @_;

    my $res = {};
    
    if ($key =~ m/^([^\d]+)(\d+)$/) {
	$res->{interface} = $1;
	$res->{index} = $2;
    } else {
	return undef;
    }

    foreach my $p (split (/,/, $data)) {
	next if $p =~ m/^\s*$/;

	if ($p =~ m/^(file|volume|cyls|heads|secs|trans|media|snapshot|cache|format|rerror|werror|backup)=(.+)$/) {
	    my ($k, $v) = ($1, $2);

	    $k = 'file' if $k eq 'volume';

	    return undef if defined $res->{$k};
	    
	    $res->{$k} = $v;
	} else {
	    if (!$res->{file} && $p !~ m/=/) {
		$res->{file} = $p;
	    } else {
		return undef;
	    }
	}
    }

    return undef if !$res->{file};

    return undef if $res->{cache} && 
	$res->{cache} !~ m/^(off|none|writethrough|writeback)$/;
    return undef if $res->{snapshot} && $res->{snapshot} !~ m/^(on|off)$/;
    return undef if $res->{cyls} && $res->{cyls} !~ m/^\d+$/;
    return undef if $res->{heads} && $res->{heads} !~ m/^\d+$/;
    return undef if $res->{secs} && $res->{secs} !~ m/^\d+$/;
    return undef if $res->{media} && $res->{media} !~ m/^(disk|cdrom)$/;
    return undef if $res->{trans} && $res->{trans} !~ m/^(none|lba|auto)$/;
    return undef if $res->{format} && $res->{format} !~ m/^(raw|cow|qcow|qcow2|vmdk|cloop)$/;
    return undef if $res->{rerror} && $res->{rerror} !~ m/^(ignore|report|stop)$/;
    return undef if $res->{werror} && $res->{werror} !~ m/^(enospc|ignore|report|stop)$/;
    return undef if $res->{backup} && $res->{backup} !~ m/^(yes|no)$/;

    if ($res->{media} && ($res->{media} eq 'cdrom')) {
	return undef if $res->{snapshot} || $res->{trans} || $res->{format};
	return undef if $res->{heads} || $res->{secs} || $res->{cyls}; 
	return undef if $res->{interface} eq 'virtio';
    }

    # rerror does not work with scsi drives
    if ($res->{rerror}) {
	return undef if $res->{interface} eq 'scsi';
    }

    return $res;
}
   
my @qemu_drive_options = qw(heads secs cyls trans media format cache snapshot rerror werror);

sub print_drive {
    my ($vmid, $drive) = @_;

    my $opts = '';
    foreach my $o (@qemu_drive_options, 'backup') {
	$opts .= ",$o=$drive->{$o}" if $drive->{$o};
    }
    
    return "$drive->{file}$opts";
}

sub print_drive_full {
    my ($self, $vmid, $drive) = @_;

    my $opts = '';
    foreach my $o (@qemu_drive_options) {
	$opts .= ",$o=$drive->{$o}" if $drive->{$o};
    }
 
    my $path;
    my $volid = $drive->{file};
    if (drive_is_cdrom ($drive)) {
	$path = get_iso_path ($self->{storecfg}, $vmid, $volid);
    } else {
	if ($volid =~ m|^/|) {
	    $path = $volid;
	} else {
	    $path = PVE::Storage::path ($self->{storecfg}, $volid);
	}
	if (!$drive->{cache} && ($path =~ m|^/dev/| || $path =~ m|\.raw$|)) {
	    $opts .= ",cache=none";
	}
    }

    my $pathinfo = $path ? "file=$path," : '';

    return "${pathinfo}if=$drive->{interface},index=$drive->{index}$opts";
}


sub drive_is_cdrom {
    my ($drive) = @_;

    return $drive && $drive->{media} && ($drive->{media} eq 'cdrom');

}
# vlanX: e1000=XX:XX:XX:XX:XX:XX,pcnet=XX:XX:XX:XX:XX:XX,...

sub parse_vlan {
    my ($data) = @_;

    my $res = {};

    foreach my $nic (split (/,/, $data)) {

	if ($nic =~ m/^(ne2k_pci|e1000|rtl8139|pcnet|virtio|ne2k_isa|i82551|i82557b|i82559er)(=([0-9a-f]{2}(:[0-9a-f]{2}){5}))?$/i) {
	    my $model = lc ($1);
	    my $mac = uc($3) || random_ether_addr ();
	    push @{$res->{nics}}, { model => $model, macaddr => $mac } ;
	} else {
	    return undef;
	}
	
    }

    return undef if !$res->{nics};

    return $res;

}

sub print_vlan {
    my $vlan = shift;

    my $res = '';
    foreach my $nic (@{$vlan->{nics}}) {
	$res .= ',' if $res;
	$res .= "$nic->{model}";
	my $mac =  $nic->{macaddr};
	$res .= "=$mac" if $mac;
    }
    return $res;
}

sub check_type {
    my ($vmid, $type, $key, $value) = @_;

    my $res;

    if ($type eq 'bool') {
	return 1 if ($value eq '1') || ($value =~ m/^(on|yes|true)$/i); 
	return 0 if ($value eq '0') || ($value =~ m/^(off|no|false)$/i); 
    } elsif ($type eq 'natural') {
	return $1 if $value =~ m/^(\d+)$/;
    } elsif ($type eq 'lang') {
	$value = lc ($value);
	return $value if -f "/usr/share/kvm/keymaps/$value";
    } elsif ($type eq 'string') {
	$value =~ s/^\"(.*)\"$/$1/;
	return $value;	
    } elsif ($type eq 'ostype') {
	return $value if $value =~ m/^(other|wxp|w2k|w2k3|w2k8|wvista|l24|l26)$/;
    } elsif ($type eq 'vgatype') {
	return $value if $value =~ m/^(std|cirrus)$/;
    } elsif ($type eq 'boot') {
	return $value if $value =~ m/^[acdn]{1,3}$/;
    } elsif ($type eq 'bootdisk') {
	foreach my $ds (disknames()) {
	    return $value if $value eq $ds;
	}
	return undef;
    } elsif ($type eq 'lockinfo') {
	return $value if $value =~ m/^(migrate|backup)$/;	
    } elsif ($type eq 'vlan') {
	return $value if parse_vlan ($value);	
    } elsif ($type eq 'drive') {
	my $drive = parse_drive ($key, $value);
	return $value if $drive;
    } elsif ($type eq 'hostpci') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m/^[a-f0-9]{2}:[a-f0-9]{2}\.[a-f0-9]$/i;
	}
	return $value;
    } elsif ($type eq 'hostusb') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m/^(\d+.(\d+|\*)|[0-9A-Fa-f]+:([0-9A-Fa-f]+|\*))$/
	}
	return $value;
    } elsif ($type eq 'parallel') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m|^/dev/parport\d+$|
	}
	return $value;	
    } elsif ($type eq 'serial') {
	my @dl = split (/,/, $value);
	foreach my $v (@dl) {
	    return undef if $v !~ m|^/dev/ttyS\d+$|
	}
	return $value;	
    } elsif ($type eq 'startdate') {
	return $value if $value eq 'now';
	return $value if $value =~ m/^\d{4}-\d{1,2}-\d{1,2}(T\d{1,2}-\d{1,2}-\d{1,2})?$/;
    }

    return undef;
}

sub lock_config {
    my ($vmid, $code, @param) = @_;

    my $filename = config_file_lock ($vmid);

    lock_file ($filename, "VM $vmid", $code, @param);
}

# flock: we use one file handle per process, so lock file
# can be called multiple times and succeeds for the same process.

my $lock_handles =  {};

sub lock_file {
    my ($filename, $text, $code, @param) = @_;

    my $timeout = 10;

    my $res;

    eval {

	local $SIG{ALRM} = sub { die "got timeout\n"; };

	alarm ($timeout);

	if (!$lock_handles->{$$}->{$filename}) {
	    $lock_handles->{$$}->{$filename} = new IO::File (">>$filename") ||
		die "can't open lock for $text '$filename' - $!\n";
	}

	if (!flock ($lock_handles->{$$}->{$filename}, LOCK_EX|LOCK_NB)) {
	    print STDERR "trying to aquire lock...";
	    if (!flock ($lock_handles->{$$}->{$filename}, LOCK_EX)) {
		print STDERR " failed\n";
		die "can't aquire lock for $text '$filename' - $!\n";
	    }
	    print STDERR " OK\n";
	}
	alarm (0);

	$res = &$code (@param);
    };

    my $err = $@;

    alarm (0);

    if ($lock_handles->{$$}->{$filename}) {
	my $fh = $lock_handles->{$$}->{$filename};
	$lock_handles->{$$}->{$filename} = undef;
	close ($fh);
    }

    if ($err) {
	$@ = $err;
	return undef;
    }

    $@ = undef;

    return $res;
}

sub config_file {
    my ($vmid) = @_;

    return "$confdir/$vmid.conf";
}

sub config_file_lock {
    my ($vmid) = @_;

    return "$lock_dir/lock-$vmid.conf";
}

sub touch_config {
    my ($vmid) = @_;

    my $conf = config_file ($vmid);
    utime undef, undef, $conf;
}

sub unlink_image {
    my ($self, $vmid, $volid) = @_;

    die "reject to unlink absolute path '$volid'"
	if $volid =~ m|^/|;
    
    my ($path, $owner) = PVE::Storage::path ($self->{storecfg}, $volid);

    die "reject to unlink '$volid' - not owned by this VM"
	if !$owner || ($owner != $vmid);

    syslog ('info', "VM $vmid deleting volume '$volid'");

    PVE::Storage::vdisk_free ($self->{storecfg}, $volid);

    touch_config ($vmid);
}

sub destroy_vm {
    my ($self, $vmid) = @_;

    my $conffile = config_file ($vmid);

    my $conf = $self->load_config ($vmid);

    check_lock ($conf);

    my $di = $conf->{diskinfo};

    # only remove disks owned by this VM 
    foreach my $ds (keys %$di) {
	next if drive_is_cdrom ($di->{$ds});

	my $volid = $di->{$ds}->{file};
	next if !$volid;
	next if $volid =~ m|^/|;

	my ($path, $owner) = PVE::Storage::path ($self->{storecfg}, $volid);
	next if !$path;
	next if !$owner || ($owner != $vmid);

	PVE::Storage::vdisk_free ($self->{storecfg}, $volid);
    }
    
    system ("rm -f '$conffile'");

    # also remove unused disk
    eval {
	my $dl = PVE::Storage::vdisk_list ($self->{storecfg}, undef, $vmid);

	eval {
	    PVE::Storage::foreach_volid ($dl, sub {
		my ($volid, $sid, $volname, $d) = @_;
		PVE::Storage::vdisk_free ($self->{storecfg}, $volid);	    
	    });
	};
	warn $@ if $@;

    };
    warn $@ if $@;
}

sub load_diskinfo {
    my ($storecfg, $vmid, $conf) = @_;

    my @disks = disknames();

    my $info = {};
    my $res = {};
    my $vollist;
    foreach my $ds (@disks) {
	next if !$conf->{$ds};

	my $di = parse_drive ($ds, $conf->{$ds});
	next if !$di;

	$res->{$ds} = $di;

	next if drive_is_cdrom ($di);

	if ($di->{file} =~ m|^/dev/.+|) {
	    $info->{$di->{file}}->{size} = PVE::Storage::file_size_info ($di->{file});
	} else {
	    push @$vollist, $di->{file};
	}
    }

    eval {
	my $dl = PVE::Storage::vdisk_list ($storecfg, undef, $vmid, $vollist);

	PVE::Storage::foreach_volid ($dl, sub {
	    my ($volid, $sid, $volname, $d) = @_;
	    $info->{$volid} = $d;
	});
    };
    warn $@ if $@;

    foreach my $ds (keys %$res) {
	my $di = $res->{$ds};

	$res->{$ds}->{disksize} = $info->{$di->{file}} ? 
	    $info->{$di->{file}}->{size} / (1024*1024) : 0;
    }

    return $res;
}

sub load_config {
    my ($self, $vmid) = @_;

    my $filename = config_file ($vmid);

    my $fh = new IO::File ($filename, "r") ||
	die "unable to read config for VM $vmid\n";
 

    return parse_config ($filename, $fh, $self->{storecfg});
}  

sub parse_config {
    my ($filename, $fh, $storecfg) = @_;

    my $res = {};

    $filename =~ m|/(\d+)\.conf$| 
	|| die "got strange filename '$filename'";

    my $vmid = $1;

    while (my $line = <$fh>) {
	
	next if $line =~ m/^\#/;

	next if $line =~ m/^\s*$/;

	if ($line =~ m/^(description|args):\s*(.*\S)\s*$/) {
	    my $key = $1;
	    my $value = $2;
	    $res->{$key} = $value;
	} elsif ($line =~ m/^([a-z][a-z_]*\d*):\s*(\S+)\s*$/) {
	    my $key = $1;
	    my $value = $2;
	    if (my $type = $confvars->{$key}) {
		if (defined ($value = check_type ($vmid, $type, $key, $value))) {

		    if ($type eq 'drive') {
			my $v = parse_drive ($key, $value);
			if (my $volid = filename_to_volume_id ($vmid, $v->{file}, $v->{media})) {
			    $v->{file} = $volid;
			    $value = print_drive ($vmid, $v);
			} else {
			    warn "vm $vmid - unable to parse value of '$key'\n";
			    next;
			}
		    }

		    $res->{$key} = $value;

		} else {
		    warn "vm $vmid - unable to parse value of '$key'\n";
		}
	    } else {
		warn "vm $vmid - unknown setting '$key'\n";
	    }
	}
    }

    # convert old smp to sockets
    if ($res->{smp} && !$res->{sockets}) {
	$res->{sockets} = $res->{smp};
    } 
    delete $res->{smp};

    my $disksize = 0;
    my $disktype = 'none';
 
    my $di = load_diskinfo ($storecfg, $vmid, $res);

    foreach my $ds (keys %$di) {
	my $size = $di->{$ds}->{disksize};
	if ($res->{bootdisk} && ($ds eq $res->{bootdisk}) && $size) {
	    $disksize = $size;
	    $disktype = $di->{$ds}->{interface};
	}
    }

    $res->{disksize} = $disksize;
    $res->{disktype} = $disktype;
    $res->{diskinfo} = $di;

    return $res;
}

sub change_config {
    my ($vmid, $settings, $unset, $skiplock) = @_;

    lock_config ($vmid, &change_config_nolock, $settings, $unset, $skiplock);

    die $@ if $@;
}

sub change_config_nolock {
    my ($vmid, $settings, $unset, $skiplock) = @_;

    my $res = {};

    check_lock ($settings) if !$skiplock;

    # we do not use 'smp' any longer
    if ($settings->{sockets}) {
	$unset->{smp} = 1; 
    } elsif ($settings->{smp}) {
	$settings->{sockets} = $settings->{smp};
	$unset->{smp} = 1;
    }

    foreach my $key (keys %$settings) {
	my $value = $settings->{$key};
	if (my $type = $confvars->{$key}) {
	    if (defined ($value = check_type ($vmid, $type, $key, $value))) {
		$res->{$key} = $value;
	    } else {
		die "unable to parse value of '$key'\n";
	    }
	} else {
	    die "unknown setting '$key'\n";
	}
    }

    my $filename = config_file ($vmid);
    my $tmpfn = "$filename.$$.tmp";

    my $fh = new IO::File ($filename, "r") ||
	die "unable to read config for VM $vmid\n";

    my $werror = "unable to write config for VM $vmid\n";

    my $out = new IO::File ($tmpfn, "w") || die $werror;

    eval {

	my $done;

	while (my $line = <$fh>) {
	
	    if (($line =~ m/^\#/) || ($line =~ m/^\s*$/)) {
		die $werror unless print $out $line;
		next;
	    }

	    if ($line =~ m/^([a-z][a-z_]*\d*):\s*(.*\S)\s*$/) {
		my $key = $1;
		my $value = $2;

		# convert 'smp' to 'sockets'
		$key = 'sockets' if $key eq 'smp';

		next if $done->{$key};

		if (defined ($res->{$key})) {
		    $value = $res->{$key};
		    delete $res->{$key};
		}
		if (!defined ($unset->{$key})) {
		    die $werror unless print $out "$key: $value\n";
		} 
		$done->{$key} = 1;
		next;
	    }

	    die "unable to parse config file: $line\n";
	}

	foreach my $key (keys %$res) {

	    if (!defined ($unset->{$key})) {
		die $werror unless print $out "$key: $res->{$key}\n";
	    }
	}
    };

    my $err = $@;

    $fh->close();

    if ($err) {
	$out->close();
	unlink $tmpfn;
	die $err;
    }

    if (!$out->close()) {
	$err = "close failed - $!\n";
	unlink $tmpfn;
	die $err;	
    }

    if (!rename($tmpfn, $filename)) {
	$err = "rename failed - $!\n";
	unlink $tmpfn;
	die $err;
    }
}

sub load_defaults { 

    my $filename = "/etc/pve/qemu-server.cfg";

    my $res = {};

    while (my ($key, $value) = each %$defvars) {
	my ($type, $default) = @$value;
	$res->{$key} = $default if defined ($default);
    }
   
    my $fh = new IO::File ($filename, "r") || return $res;

    while (my $line = <$fh>) {
	
	next if $line =~ m/^\#/;

	next if $line =~ m/^\s*$/;

	if ($line =~ m/^([a-z][a-z_]*):\s*(\S+)\s*$/) {
	    my $key = $1;
	    my $value = $2;
	    my ($type, $default);

	    ($type, $default) = @{$defvars->{$key}} if defined $defvars->{$key};

	    if ($type) {
		if (defined ($value = check_type (undef, $type, $key, $value))) {
		    $res->{$key} = $value;
		} else {
		    warn "load_defaults - unable to parse value of '$key'\n";
		}
	    } else {
		warn "load_defaults - unknown setting '$key'\n";
	    }
	}
    }

    return $res;
}

sub config_list {

    my $fd = IO::Dir->new ($confdir);
    my $res = {};

    return $res if !$fd;

    while (defined(my $de = $fd->read)) { 
	if ($de =~ m/^(\d+)\.conf$/) {
	    $res->{$1}->{exists} = 1;
	}
    }
    return $res;
}

sub check_lock {
    my ($conf) = @_;

    die "VM is locked ($conf->{lock})\n" if $conf->{lock};
}

sub check_cmdline {
    my ($pidfile, $pid) = @_;

    my $fh = IO::File->new ("/proc/$pid/cmdline", "r");
    if (defined ($fh)) {
	my $line = <$fh>;
	$fh->close;
	return undef if !$line;
	my @param = split (/\0/, $line);

	my $cmd = $param[0];
	return if !$cmd || ($cmd !~ m|kvm$|);

	for (my $i = 0; $i < scalar (@param); $i++) {
	    my $p = $param[$i];
	    next if !$p;
	    if (($p eq '-pidfile') || ($p eq '--pidfile')) {
		my $p = $param[$i+1];
		return 1 if $p && ($p eq $pidfile);
		return undef;
	    }
	}
    }
    return undef;
}

sub check_running {
    my ($vmid) = @_;

    my $filename = config_file ($vmid);

    die "unable to find configuration file for VM $vmid - no such machine\n"
	if ! -f $filename;

    my $pidfile = pidfile_name ($vmid);

    if (my $fd = IO::File->new ("<$pidfile")) {
	my $st = stat ($fd);
	my $line = <$fd>;
	close ($fd);

	my $mtime = $st->mtime;
	if ($mtime > time()) {
	    warn "file '$filename' modified in future\n";
	}

	if ($line =~ m/^(\d+)$/) {
	    my $pid = $1;

	    return $pid if ((-d "/proc/$pid") && check_cmdline ($pidfile, $pid));
	}
    }

    return undef;
}

sub vzlist {
    
    my $vzlist = config_list();

    my $fd = IO::Dir->new ($var_run_tmpdir) || return $vzlist;

    while (defined(my $de = $fd->read)) { 
	next if $de !~ m/^(\d+)\.pid$/;
	my $vmid = $1;
	next if !defined ($vzlist->{$vmid});
	if (my $pid = check_running ($vmid)) {
	    $vzlist->{$vmid}->{pid} = $pid;
	}
    }

    return $vzlist;
}

sub config_to_command {
    my ($self, $vmid, $conf, $defaults, $migrate_uri) = @_;

    my $cmd = [];

    my $kvmver = kvm_user_version();
    my $vernum = 0; # unknown
    if ($kvmver =~ m/^(\d+)\.(\d+)\.(\d+)$/) {
	$vernum = $1*1000000+$2*1000+$3;
    } else {
	die "unable to parse kvm version '$kvmver'\n";
    }

    my $have_ovz = -f '/proc/vz/vestat';

    push @$cmd, '/usr/bin/kvm';

    my $use_virtio = 0;

    my $socket = monitor_socket ($vmid);
    push @$cmd, '-monitor', "unix:$socket,server,nowait";

    $socket = vnc_socket ($vmid);
    push @$cmd,  '-vnc', "unix:$socket,password";

    push @$cmd, '-pidfile' , pidfile_name ($vmid);
 
    push @$cmd, '-daemonize';

    push @$cmd, '-incoming', $migrate_uri if $migrate_uri;

    # enable absolute mouse coordinates (needed by vnc)
    my $tablet = defined ($conf->{tablet}) ? $conf->{tablet} : $defaults->{tablet};
    push @$cmd, '-usbdevice', 'tablet' if $tablet;

    # host pci devices
    if (my $pcidl = $conf->{hostpci}) {
	my @dl = split (/,/, $pcidl);
	foreach my $dev (@dl) {
	    next if !$dev;
	    if ($vernum < 13000) {
		push @$cmd, '-pcidevice', "host=$dev";
	    } else {
		push @$cmd, '-device', "pci-assign,host=$dev";
	    }
	}
    }
    # host usb devices
    if (my $usbdl = $conf->{hostusb}) {
	my @dl = split (/,/, $usbdl);
	foreach my $dev (@dl) {
	    push @$cmd, '-usbdevice', "host:$dev" if $dev;
	}
    }

    # serial devices
    if (my $serdl = $conf->{serial}) {
	my @dl = split (/,/, $serdl);
	foreach my $dev (@dl) {
	    next if !$dev;
	    if (-c $dev) {
		push @$cmd, '-serial', "$dev";
	    }
	}
    }

    # parallel devices
    if (my $pardl = $conf->{parallel}) {
	my @dl = split (/,/, $pardl);
	foreach my $dev (@dl) {
	    next if !$dev;
	    if (-c $dev) {
		push @$cmd, '-parallel', "$dev";
	    }
	}
    }

    my $vmname = $conf->{name} || "vm$vmid";

    push @$cmd, '-name', $vmname;
    
    my $sockets = 1;
    $sockets = $conf->{smp} if $conf->{smp}; # old style - no longer used
    $sockets = $conf->{sockets} if  $conf->{sockets};

    my $cores = $conf->{cores} || 1;

    my $boot_opt;

    if ($vernum < 11000) { # 0.9.X and 0.10.X

	push @$cmd, '-smp', $sockets*$cores;

	$boot_opt = $conf->{boot};
    } else {

	push @$cmd, '-smp', "sockets=$sockets,cores=$cores";

	$boot_opt = "menu=on";
	if ($conf->{boot}) {
	    $boot_opt .= ",order=$conf->{boot}";
	}

	push @$cmd, '-nodefaults' if $vernum >= 12003;

    }

    push @$cmd, '-boot', $boot_opt if $boot_opt;

    push @$cmd, '-no-acpi' if defined ($conf->{acpi}) && $conf->{acpi} == 0;

    push @$cmd, '-no-reboot' if  defined ($conf->{reboot}) && $conf->{reboot} == 0;

    my $vga = $conf->{vga} || $defaults->{vga};
    push @$cmd, '-vga', $vga if $vga; # for kvm 77 and later

    # time drift fix
    my $tdf = defined ($conf->{tdf}) ? $conf->{tdf} : $defaults->{tdf};
    push @$cmd, '-tdf' if $tdf;

    my $nokvm = defined ($conf->{kvm}) && $conf->{kvm} == 0 ? 1 : 0;

    if (my $ost = $conf->{ostype}) {
	# other, wxp, w2k, w2k3, w2k8, wvista, l24, l26

	if ($ost =~ m/^w/) { # windows
	    push @$cmd, '-localtime' if !defined ($conf->{localtime});

	    # use rtc-td-hack when acpi is enabled
	    if (!(defined ($conf->{acpi}) && $conf->{acpi} == 0)) {
		push @$cmd, '-rtc-td-hack';
	    }
	}

	# -tdf ?
	# -no-acpi 
	# -no-kvm 
	# -win2k-hack ?
    }

    push @$cmd, '-no-kvm' if $nokvm;

    push @$cmd, '-localtime' if $conf->{localtime};

    push @$cmd, '-startdate', $conf->{startdate} if $conf->{startdate};

    push @$cmd, '-S' if $conf->{freeze};

    # set keyboard layout
    my $kb = $conf->{keyboard} || $defaults->{keyboard};
    push @$cmd, '-k', $kb if $kb;

    # enable sound
    #my $soundhw = $conf->{soundhw} || $defaults->{soundhw};
    #push @$cmd, '-soundhw', 'es1370';
    #push @$cmd, '-soundhw', $soundhw if $soundhw;

    my $di = $conf->{diskinfo};
    foreach my $ds (sort keys %$di) {
	$use_virtio = 1 if $ds =~ m/^virtio/;
	my $drive = $self->print_drive_full ($vmid, $di->{$ds});
	$drive .= ",boot=on" if $conf->{bootdisk} && ($conf->{bootdisk} eq $ds);
	push @$cmd, '-drive', $drive;
    }

    if ($conf->{memory}) {
	push @$cmd, '-m', $conf->{memory};
    } else {
	push @$cmd, '-m', $defaults->{memory} || $default_memory;
    }

    my $foundnet = 0;

    foreach my $k (sort keys %$conf) {
	next if $k !~ m/^vlan(\d+)$/;
	my $i = int ($1);

	die "got strange vlan id '$i'\n" if $i >= ${MAX_VLANS};

	if ($conf->{"vlan$i"} && (my $vlan = parse_vlan ($conf->{"vlan$i"}))) {

	    $foundnet = 1;

	    my $baseifname = "tap${vmid}i$i";

	    if ($vernum < 13000) {
		push @$cmd, '-net', "tap,vlan=$i,ifname=$baseifname,script=/var/lib/qemu-server/bridge-vlan";
	    }

	    my $j = 0;
	    foreach my $nic (@{$vlan->{nics}}) {
		my $ifname = $baseifname;
		my $device = $nic->{model};
		my $vhostparam = '';
		if ($nic->{model} eq 'virtio') {
		    $use_virtio = 1;
		    $device = 'virtio-net-pci';
		    $vhostparam = ',vhost=on' if $kernel_has_vhost_net;
		};
		if ($vernum < 13000) {
		    push @$cmd, '-net', "nic,vlan=$i,model=$nic->{model},macaddr=$nic->{macaddr}";
		} else {
		    $ifname = "${ifname}d$j";
		    push @$cmd, '-netdev', "type=tap,id=${k}d$j,ifname=${ifname},script=/var/lib/qemu-server/bridge-vlan$vhostparam";
		    # qemu 0.15 always try to boot from network - we disable that by
		    # not loading the pxe rom file
		    my $extra = (($vernum >= 15000) && (!$conf->{boot} || ($conf->{boot} !~ m/n/))) ?
			"romfile=," : '';
		    push @$cmd, '-device', "$device,${extra}mac=$nic->{macaddr},netdev=${k}d$j";
		}

		# kvm uses TUNSETIFF ioctl, and that limits ifname length
		die "interface name '$ifname' is too long (max 15 character)\n" 
		    if length($ifname) >= 16;

		$j = $j + 1;
	    }

	}
    } 

    if ($conf->{"vlanu"} && (my $vlan = parse_vlan ($conf->{"vlanu"}))) {

	$foundnet = 1;

	push @$cmd, '-net', "user,vlan=${MAX_VLANS},hostname=$vmname";

	foreach my $nic (@{$vlan->{nics}}) {
	    $use_virtio = 1 if $nic->{model} eq 'virtio';
	    push @$cmd, '-net', "nic,vlan=${MAX_VLANS},model=$nic->{model},macaddr=$nic->{macaddr}";
	}
    }
    
    push @$cmd, '-net', 'none' if !$foundnet;

    # hack: virtio with fairsched is unreliable, so we do not use fairsched
    # when the VM uses virtio devices.
    if (!$use_virtio && $have_ovz) { 

	push @$cmd, '-id', $vmid;
     
	my $cpuunits = defined ($conf->{cpuunits}) ? 
	    $conf->{cpuunits} : $defaults->{cpuunits};

	push @$cmd, '-cpuunits', $cpuunits if $cpuunits;

	# fixme: cpulimit is currently ignored
	#push @$cmd, '-cpulimit', $conf->{cpulimit} if $conf->{cpulimit};
    }

    # add custom args
    if ($conf->{args}) {
	my $aa = split_args ($conf->{args});
	push @$cmd, @$aa;
    }

    return $cmd;
}
 
sub vnc_socket {
    my ($vmid) = @_;
    return "${var_run_tmpdir}/$vmid.vnc";
}

sub monitor_socket {
    my ($vmid) = @_;
    return "${var_run_tmpdir}/$vmid.mon";
}

sub pidfile_name {
    my ($vmid) = @_;
    return "${var_run_tmpdir}/$vmid.pid";
}

sub random_ether_addr {

    my $rand = Digest::SHA1::sha1_hex (rand(), time());

    my $mac = '';
    for (my $i = 0; $i < 6; $i++) {
	my $ss = hex (substr ($rand, $i*2, 2));
	if (!$i) {
	    $ss &= 0xfe; # clear multicast
	    $ss |= 2; # set local id
	}
	$ss = sprintf ("%02X", $ss);

	if (!$i) {
	    $mac .= "$ss";
	} else {
	    $mac .= ":$ss";
	}
    }
    
    return $mac;
}

sub next_migrate_port {

    for (my $p = 60000; $p < 60010; $p++) {

	my $sock = IO::Socket::INET->new (Listen => 5,
					  LocalAddr => 'localhost',
					  LocalPort => $p,
					  ReuseAddr => 1,
					  Proto     => 0);

	if ($sock) {
	    close ($sock);
	    return $p;
	}
    }

    die "unable to find free migration port";
}

sub activate_volumes {
    my ($self, $conf)  = @_;

    my $di = $conf->{diskinfo};

    my $vollist = [];
    foreach my $ds (keys %$di) {
	my $volid = $di->{$ds}->{file};
	eval { 
	    PVE::Storage::parse_volume_id ($volid);
	    push @$vollist, $volid;
	};
    }

    PVE::Storage::activate_volumes ($self->{storecfg}, $vollist);
}

sub vm_start {
    my ($self, $vmid, $statefile, $skiplock) = @_;

    lock_config ($vmid, sub {
	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	mkpath $var_run_tmpdir;
    
	if (check_running ($vmid)) {
	    my $msg = "VM $vmid already running - start failed\n" ;
	    syslog ('err', $msg);
	    die $msg;
	} else {
	    syslog ('info', "VM $vmid start");
	}

	my $migrate_uri;
	my $migrate_port = 0;

	if ($statefile) {
	    if ($statefile eq 'tcp') {
		$migrate_port = next_migrate_port();
		$migrate_uri = "tcp:localhost:${migrate_port}";
	    } else {
		if (-f $statefile) {
		    $migrate_uri = "exec:cat $statefile";
		} else {
		    warn "state file '$statefile' does not exist - doing normal startup\n";
		}
	    }
	}

	my $defaults = load_defaults();

	my $cmd = $self->config_to_command ($vmid, $conf, $defaults, $migrate_uri);
	# host pci devices
	if (my $pcidl = $conf->{hostpci}) {
	    my @dl = split (/,/, $pcidl);
	    foreach my $dev (@dl) {
		$dev = lc($dev);
		my $info = pci_device_info("0000:$dev");
		die "no pci device info for device '$dev'\n" if !$info;
		die "can't unbind pci device '$dev'\n" if !pci_dev_bind_to_stub($info);
		die "can't reset pci device '$dev'\n" if !pci_dev_reset($info);
	    }
	}

	$self->activate_volumes ($conf);

	eval  { PVE::Storage::run_command ($cmd, timeout => $migrate_uri ? undef : 30); };

	my $err = $@;

	if ($err) {
	    my $msg = "start failed: $err";
	    syslog ('err', "VM $vmid $msg");
	    die $msg;
	}

	if ($statefile) {

	    if ($statefile eq 'tcp') {
		print "migration listens on port $migrate_port\n";
	    } else {
		unlink $statefile;
		# fixme: send resume - is that necessary ?
		eval  { $self->vm_monitor_command ($vmid, "cont", 1) };
	    }
	}

	if (my $migrate_speed = 
	    $conf->{migrate_speed} || $defaults->{migrate_speed}) {
	    my $cmd = "migrate_set_speed ${migrate_speed}m";
	    eval { $self->vm_monitor_command ($vmid, $cmd, 1); };
	}

	if (my $migrate_downtime = 
	    $conf->{migrate_downtime} || $defaults->{migrate_downtime}) {
	    my $cmd = "migrate_set_downtime ${migrate_downtime}";
	    eval { $self->vm_monitor_command ($vmid, $cmd, 1); };
	}
    });

    die $@ if $@;
}

sub __read_avail {
    my ($fh, $timeout) = @_;

    my $sel = new IO::Select;
    $sel->add ($fh);

    my $res = '';
    my $buf;

    my @ready;
    while (scalar (@ready = $sel->can_read ($timeout))) {
	my $count;
	if ($count = $fh->sysread ($buf, 8192)) {
	    if ($buf =~ /^(.*)\(qemu\) $/s) {
		$res .= $1;
		last;
	    } else {
		$res .= $buf;
	    }
	} else {
	    if (!defined ($count)) {
		die "$!\n";
	    }
	    last;
	}
    }

    die "monitor read timeout\n" if !scalar (@ready);

    return $res;
}

sub vm_monitor_command {
    my ($self, $vmid, $cmdstr, $nolog) = @_;

    my $res;

    syslog ("info", "VM $vmid monitor command '$cmdstr'") if !$nolog;

    eval {
	die "VM not running\n"  if !check_running ($vmid);

	my $sname = monitor_socket ($vmid);

	my $sock = IO::Socket::UNIX->new ( Peer => $sname ) ||
	    die "unable to connect to VM $vmid socket - $!\n";

	my $timeout = 3;

	# hack: migrate sometime blocks the monitor (when migrate_downtime 
	# is set)
	if ($cmdstr =~ m/^(info\s+migrate|migrate\s)/) {
	    $timeout = 60*60; # 1 hour
	}

	# read banner;
	my $data = __read_avail ($sock, $timeout);
	
	if ($data !~ m/^QEMU\s+(\S+)\s+monitor\s/) {
	    die "got unexpected qemu monitor banner\n";
	}

	my $sel = new IO::Select;
	$sel->add ($sock);

	if (!scalar (my @ready = $sel->can_write ($timeout))) {
	    die "monitor write error - timeout";
	}

	my $fullcmd = "$cmdstr\r";

	my $b;
	if (!($b = $sock->syswrite ($fullcmd)) || ($b != length ($fullcmd))) {
	    die "monitor write error - $!";
	}

	return if ($cmdstr eq 'q') || ($cmdstr eq 'quit');

	$timeout = 20; 

	if ($cmdstr =~ m/^(info\s+migrate|migrate\s)/) {
	    $timeout = 60*60; # 1 hour
	} elsif ($cmdstr =~ m/^(eject|change)/) {
	    $timeout = 60; # note: cdrom mount command is slow
	}
	if ($res = __read_avail ($sock, $timeout)) {
  
	    my @lines = split ("\r?\n", $res);

	    shift @lines if $lines[0] !~ m/^unknown command/; # skip echo
	    
	    $res = join ("\n", @lines);
	    $res .= "\n";
	}
    };

    my $err = $@;

    if ($err) {
	syslog ("err", "VM $vmid monitor command failed - $err");
	die $err;
    }

    return $res;
}

sub vm_vncticket {
    my ($self, $vmid, $ticket) = @_;

    # generate ticket, olny first 8 character used by vnc
    $ticket = Digest::SHA1::sha1_base64 (rand(), time()) if !$ticket;

    $self->vm_monitor_command ($vmid, "change vnc ticket $ticket", 1);

    return $ticket;
}

sub vm_commandline {
    my ($self, $vmid) = @_;

    my $conf = $self->load_config ($vmid);

    my $defaults = load_defaults();

    my $cmd = $self->config_to_command ($vmid, $conf, $defaults);

    return join (' ', @$cmd);
}

sub vm_reset {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	syslog ("info", "VM $vmid sending 'reset'");

	$self->vm_monitor_command ($vmid, "system_reset", 1);
    });
    
    die $@ if $@;
}

sub vm_wait {
    my ($self, $vmid, $timeout, $nolog) = @_;

    my $pid = check_running ($vmid);

    syslog ("info", "VM $vmid waiting until stopped") if !$nolog;

    my $count = 0;
    while ((!$timeout || ($count < $timeout)) && check_running ($vmid)) {
	$count++;
	sleep 1;
    }

    if (check_running ($vmid)) {
	syslog ('info', "VM $vmid wait failed - got timeout");
	die "wait failed - got timeout\n";
    }
}

sub vm_shutdown {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	syslog ("info", "VM $vmid sending 'shutdown'");

	$self->vm_monitor_command ($vmid, "system_powerdown", 1);
    });

    die $@ if $@;
}

sub vm_stop {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $pid = check_running ($vmid);

	if (!$pid) {
	    syslog ('info', "VM $vmid already stopped");
	    return;
	}

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;
    
	syslog ("info", "VM $vmid stopping");

	eval { $self->vm_monitor_command ($vmid, "quit", 1); };

	my $err = $@;

	if (!$err) {
	    # wait some time
	    my $timeout = 50; # fixme: how long?

	    my $count = 0;
	    while (($count < $timeout) && check_running ($vmid)) {
		$count++;
		sleep 1;
	    }

	    if ($count >= $timeout) {
		syslog ('info', "VM $vmid still running - terminating now with SIGTERM");
		kill 15, $pid;
	    }
	} else {
	    syslog ('info', "VM $vmid quit failed - terminating now with SIGTERM");
	    kill 15, $pid;
	}

	# wait again
	my $timeout = 10;

	my $count = 0;
	while (($count < $timeout) && check_running ($vmid)) {
	    $count++;
	    sleep 1;
	}

	if ($count >= $timeout) {
	    syslog ('info', "VM $vmid still running - terminating now with SIGKILL\n");
	    kill 9, $pid;
	}

	fairsched_rmnod ($vmid); # try to destroy group
    });

    die $@ if $@;
}

sub vm_suspend {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	syslog ("info", "VM $vmid suspend");

	$self->vm_monitor_command ($vmid, "stop", 1);
    });

    die $@ if $@;
}

sub vm_resume {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	syslog ("info", "VM $vmid resume");

	$self->vm_monitor_command ($vmid, "cont", 1);
    });

    die $@ if $@;
}

sub vm_cad {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	syslog ("info", "VM $vmid sending cntl-alt-delete");
    
	$self->vm_monitor_command ($vmid, "sendkey ctrl-alt-delete", 1);
    });

    die $@ if $@;
}

sub vm_cdrom {
    my ($self, $vmid, $dev, $volid) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf);

	my $disk;
	if (($disk = $conf->{diskinfo}->{$dev}) && drive_is_cdrom ($disk)) {
	    $disk->{file} = $volid;
	    my $newdrive =  print_drive ($vmid, $disk);
	    my $settings = { $dev => $newdrive };
	    change_config_nolock  ($vmid, $settings);
	} else {
	    die "no such cdrom device '$dev'\n";
	}

	return if !check_running ($vmid); # do nothing

	$conf = $self->load_config ($vmid);# reload

	my $qdn;
	eval {
	    if (!valid_drivename ($dev)) {
		die "cdrom change failed - invalid device name '$dev'\n";
	    }

	    if ($volid eq 'eject' || $volid eq 'none') {
		syslog ("info", "VM $vmid eject cdrom '$dev'");
	    } else {
		syslog ("info", "VM $vmid change cdrom '$dev' to '$volid'");
	    }

	    $qdn = qemu_drive_name ($dev, 'cdrom');
	};

	if (my $err = $@) {
	    syslog ('err', "VM $vmid $err");
	    die $err;
	}

	if ($volid eq 'eject' || $volid eq 'none') {
	    $self->vm_monitor_command ($vmid, "eject $qdn", 1);
	} else {
	    my $path = get_iso_path ($self->{storecfg}, $vmid, $volid);

	    $self->vm_monitor_command ($vmid, "eject $qdn", 1);
	    $self->vm_monitor_command ($vmid, "change $qdn \"$path\"", 1);
	}
    });

    die $@ if $@;
}

sub vm_destroy {
    my ($self, $vmid, $skiplock) = @_;

    lock_config ($vmid, sub {

	my $conf = $self->load_config ($vmid);

	check_lock ($conf) if !$skiplock;

	syslog ("info", "VM $vmid destroy called (removing all data)");

	eval {
	    if (!check_running($vmid)) {
		fairsched_rmnod ($vmid); # try to destroy group
		$self->destroy_vm ($vmid);
	    } else {
		die "VM is running\n";
	    }
	};

	my $err = $@;

	if ($err) {
	    syslog ("err", "VM $vmid destroy failed - $err");
	    die $err;
	}
    });

    die $@ if $@;
}

sub vm_startall {
    my ($self) = @_;

    my $vzlist = vzlist();

    foreach my $vmid (keys %$vzlist) {
	next if $vzlist->{$vmid}->{pid}; # already running

	eval {
	    my $conf =  $self->load_config ($vmid);
	    if ($conf->{onboot}) {
		print STDERR "Starting Qemu VM $vmid\n";
		$self->vm_start ($vmid);
	    }
	};
	print STDERR $@ if $@;
    }
}

sub vm_stopall {
    my ($self, $timeout) = @_;

    $timeout = 3*60 if !$timeout;

    my $vzlist = vzlist();
    my $count = 0;
    foreach my $vmid (keys %$vzlist) {
	next if !$vzlist->{$vmid}->{pid};
	$count++;
    }

    if ($count) {

	my $msg = "Stopping Qemu Server - sending shutdown requests to all VMs\n";
	syslog ('info', $msg);
	print STDERR $msg;

	foreach my $vmid (keys %$vzlist) {
	    next if !$vzlist->{$vmid}->{pid};
	    eval { $self->vm_shutdown ($vmid, 1); };
	    print STDERR $@ if $@;
	}

	my $wt = 5;
	my $maxtries = int (($timeout + $wt -1)/$wt);
	my $try = 0;
	while (($try < $maxtries) && $count) {
	    $try++;
	    sleep $wt;

	    $vzlist = vzlist();
	    $count = 0;
	    foreach my $vmid (keys %$vzlist) {
		next if !$vzlist->{$vmid}->{pid};
		$count++;
	    }
	    last if !$count;
	}

	return if !$count;

	foreach my $vmid (keys %$vzlist) {
	    next if !$vzlist->{$vmid}->{pid};
	
	    $msg = "VM $vmid still running - sending stop now\n";
	    syslog ('info', $msg);
	    print $msg;

	    eval { $self->vm_monitor_command ($vmid, "quit", 1); };
	    print STDERR $@ if $@;

	}

	$timeout = 30;
	$maxtries = int (($timeout + $wt -1)/$wt);
	$try = 0;
	while (($try < $maxtries) && $count) {
	    $try++;
	    sleep $wt;

	    $vzlist = vzlist();
	    $count = 0;
	    foreach my $vmid (keys %$vzlist) {
		next if !$vzlist->{$vmid}->{pid};
		$count++;
	    }
	    last if !$count;
	}

	return if !$count;

	foreach my $vmid (keys %$vzlist) {
	    next if !$vzlist->{$vmid}->{pid};
	
	    $msg = "VM $vmid still running - terminating now with SIGTERM\n";
	    syslog ('info', $msg);
	    print $msg;
	    kill 15, $vzlist->{$vmid}->{pid};
	}

	# this is called by system shotdown scripts, so remaining
	# processes gets killed anyways (no need to send kill -9 here)

	$msg = "Qemu Server stopped\n";
	syslog ('info', $msg);
	print STDERR $msg;
    }
}

# pci helpers

sub file_read_firstline {
    my ($filename) = @_;

    my $fh = IO::File->new ($filename, "r");
    return undef if !$fh;
    my $res = <$fh>;
    chomp $res;
    $fh->close;
    return $res;
}

sub file_write {
    my ($filename, $buf) = @_;

    my $fh = IO::File->new ($filename, "w");
    return undef if !$fh;

    my $res = print $fh $buf;

    $fh->close();

    return $res;
}

sub pci_device_info {
    my ($name) = @_;

    my $res;

    return undef if $name !~ m/^([a-f0-9]{4}):([a-f0-9]{2}):([a-f0-9]{2})\.([a-f0-9])$/;
    my ($domain, $bus, $slot, $func) = ($1, $2, $3, $4);

    my $irq = file_read_firstline("$pcisysfs/devices/$name/irq");
    return undef if !defined($irq) || $irq !~ m/^\d+$/;

    my $vendor = file_read_firstline("$pcisysfs/devices/$name/vendor");
    return undef if !defined($vendor) || $vendor !~ s/^0x//;

    my $product = file_read_firstline("$pcisysfs/devices/$name/device");
    return undef if !defined($product) || $product !~ s/^0x//;

    $res = {
	name => $name,
	vendor => $vendor,
	product => $product,
	domain => $domain,
	bus => $bus,
	slot => $slot,
	func => $func,
	irq => $irq,
	has_fl_reset => -f "$pcisysfs/devices/$name/reset" || 0,
    };

    return $res;
}

sub pci_dev_reset {
    my ($dev) = @_;

    my $name = $dev->{name};

    my $fn = "$pcisysfs/devices/$name/reset";

    return file_write ($fn, "1");
}

sub pci_dev_bind_to_stub {
    my ($dev) = @_;

    my $name = $dev->{name};

    my $testdir = "$pcisysfs/drivers/pci-stub/$name";
    return 1 if -d $testdir;

    my $data = "$dev->{vendor} $dev->{product}";
    return undef if !file_write ("$pcisysfs/drivers/pci-stub/new_id", $data);

    my $fn = "$pcisysfs/devices/$name/driver/unbind";
    if (!file_write ($fn, $name)) {
	return undef if -f $fn;
    }

    $fn = "$pcisysfs/drivers/pci-stub/bind";
    if (! -d $testdir) {
	return undef if !file_write ($fn, $name);
    }

    return -d $testdir;
}

1;
